#include <iterator>

#include "constraint.hpp"
#include "parser-classes.hpp"
#include "log.hpp"
#include "resmes.hpp"

using namespace std;

int conaux::max_var_amount;
map<string, bdd> conaux::varmap;
vector<string> conaux::varnames;

ostream &operator<<(ostream &os, const bddStat &s)
{
   os << "   total number of new nodes ever produced: " << s.produced << endl;
   os << "   currently allocated number of bdd nodes: " << s.nodenum << endl;
   os << "   user defined maximum number of bdd nodes: " << s.maxnodenum << endl;
   os << "   number of currently free nodes: " << s.freenodes << endl;
   os << "   minimum number of nodes that should be left after a garbage collection: " << s.minfreenodes << endl;
   os << "   number of defined bdd variables: " << s.varnum << endl;
   os << "   number of entries in the internal caches: " << s.cachesize << endl;
   os << "   number of garbage collections done until now: " << s.gbcnum << endl;
   return os;
}

ostream &operator<<(ostream &os, const bddCacheStat &s)
{
   os << "   accesses to the unique node table: " << s.uniqueAccess << endl;
   os << "   iterations through the cache chains in the unique node table: " << s.uniqueChain << endl;
   os << "   entries actually found in the unique node table: " << s.uniqueHit << endl;
   os << "   entries not found in the unique node table: " << s.uniqueMiss << endl;
   os << "   entries found in the operator caches: " << s.opHit << endl;
   os << "   entries not found in the operator caches: " << s.opMiss << endl;
   os << "   number of variable swaps in reordering: " << s.swapCount << endl;
   return os;
}

void conaux::init(const string &constraint_settings)
{
   lout(L_STATUS) << "Initialising constraints..." << endl;
   istringstream iss(constraint_settings);
   int nodesize, cachesize;
   iss >> nodesize >> cachesize >> max_var_amount;
   bdd_init(nodesize, cachesize);
   bdd_setvarnum(max_var_amount);
}

void conaux::finalise()
{
   bddStat s;
   bdd_stats(&s);

   lout(L_RES) << endl << endl;

   // Ordinary stats
   lout(L_RES) << "BuDDy statistics" << endl 
	       << "----------------" << endl
	       << s << endl;
   
   // Finalise the BDD table
   bdd_done();
}

string conaux::get_varname(const int num)
{
   return varnames[num];
}

bdd conaux::get_variable(const string &name)
{
   map<string, bdd>::iterator it = varmap.find(name);

   // That variable was already created
   if(it != varmap.end())
      return it->second;

   int varnum = varnames.size(); // Variable numbering is zero-based
   
   if(varnum >= max_var_amount)
   {
      max_var_amount *= 2;
      bdd_setvarnum(max_var_amount);
   }

   varnames.push_back(name);

   return varmap[name] = bdd_ithvar(varnum);
}

void conaux::allsathandler(char *varset, int size)
{
   vector< pair<string, bool> > ans;
   for(int i = 0; i < size; i++)
   {
      switch(varset[i])
      {
	 case 0:
	    ans.push_back(pair<string, bool>(get_varname(i), false));
	    break;
	    
	 case 1:
	    ans.push_back(pair<string, bool>(get_varname(i), true));
	    break;

	 default:
	    break;
      }
   }

   lout(L_REPORT) << "{ ";
   for(size_t i = 0; i < ans.size(); i++)
   {
      if(i != 0) lout(L_REPORT) << ", ";
      lout(L_REPORT) << ans[i].first << " = " << ans[i].second;
   }
   lout(L_REPORT) << " }" << endl;
}

void conaux::report_answer(const con_type &con)
{
   bdd_allsat(con.constraint, conaux::allsathandler);
}

con_type conaux::resolve_sat(const class c_term *t)
{
   con_type ret;
   set<string> cvars;
   ret = build_tree(t->arglist[0], cvars);
   ret.cvars = cvars;
   return ret;
}

con_type conaux::build_tree(const class c_term *term, set<string> &cvars)
{
   if(term->type == T_FUNCTOR) // A functor (and, or, xor, not, exists)
   {
      if(term->name == "not")
	 return !build_tree(term->arglist[0], cvars);
      else if(term->name == "and")
	 return build_tree(term->arglist[0], cvars) & build_tree(term->arglist[1], cvars);
      else if(term->name == "or")
	 return build_tree(term->arglist[0], cvars) | build_tree(term->arglist[1], cvars);
      else if(term->name == "xor")
	 return build_tree(term->arglist[0], cvars) ^ build_tree(term->arglist[1], cvars);
      else
	 throw sat_exception(); // We had an unrecognised functor in a sat expression
   }
   else if(term->type == T_VARIABLE)
   {
      cvars.insert(term->name);
      return con_type(term->name); // Creates if it's a new variable
   }
   else
      throw sat_exception(); // We had a list or something in a sat expression
}

void conaux::reindex(con_type &con)
{
   int size = con.cvars.size();
   int *invec = new int[size];
   int *outvec = new int[size];
   int i = 0;
   set<string>::iterator it = con.cvars.begin();
   set<string> new_cvars;
   while(it != con.cvars.end())
   {
      ostringstream oss;
      oss << *it << "#" << c_rule::global_variable_index;
      new_cvars.insert(oss.str());
      invec[i] = bdd_var(conaux::get_variable(*it));
      outvec[i] = bdd_var(conaux::get_variable(oss.str()));
      ++it;
      i++;
   }
   bddPair *p = bdd_newpair();
   bdd_setpairs(p, invec, outvec, size);
   con.constraint = bdd_replace(con.constraint, p);
   bdd_freepair(p);
   delete[] invec;
   delete[] outvec;
   con.cvars = new_cvars;
}

bool conaux::subsumes(const con_type &large, const con_type &small)
{
  lout(L_DBG) << "subsumes?:" << endl
	      << bddtable << large.constraint << endl
	      << bddtable << small.constraint << endl;
  return ((!large.constraint) & small.constraint) == bddfalse;
}

con_type::con_type(bool val)
{
   if(val == true)
      constraint = bddtrue;
   else
      constraint = bddfalse;
}

con_type::con_type(const string &var)
{
   constraint = conaux::get_variable(var);
   cvars.insert(var);
}

con_type conaux::relate(const string &var1, const string &var2)
{
   con_type ret;
   ret.constraint = bdd_biimp(conaux::get_variable(var1), conaux::get_variable(var2));
   ret.cvars.insert(var1);
   ret.cvars.insert(var2);
   return ret;
}

con_type con_type::operator&(const con_type &rhs) const
{
   con_type ret;
   resources["conop"].start();
   ret.constraint = constraint & rhs.constraint;
   ret.cvars.insert(cvars.begin(), cvars.end());
   ret.cvars.insert(rhs.cvars.begin(), rhs.cvars.end());
   resources["conop"].stop();
   return ret;
}

con_type con_type::operator!() const
{
   con_type ret;
   resources["conop"].start();
   ret.constraint = !constraint;
   ret.cvars = cvars;
   resources["conop"].stop();
   return ret;
}

con_type con_type::operator|(const con_type &rhs) const
{
   con_type ret;
   resources["conop"].start();
   ret.constraint = constraint | rhs.constraint;
   ret.cvars.insert(cvars.begin(), cvars.end());
   ret.cvars.insert(rhs.cvars.begin(), rhs.cvars.end());
   resources["conop"].stop();
   return ret;
}

con_type con_type::operator^(const con_type &rhs) const
{
   con_type ret;
   resources["conop"].start();
   ret.constraint = constraint | rhs.constraint;
   ret.cvars.insert(cvars.begin(), cvars.end());
   ret.cvars.insert(rhs.cvars.begin(), rhs.cvars.end());
   resources["conop"].stop();
   return ret;
}

bool con_type::is_false() const
{
   return constraint == bddfalse;
}

bool con_type::is_true() const
{
   return constraint == bddtrue;
}

void con_type::project(const set<string> &to_keep)
{
   con_type ret;
   set<string> to_remove;
   
   set_difference(cvars.begin(), cvars.end(), to_keep.begin(), to_keep.end(), inserter(to_remove, to_remove.end()));

   bdd bdd_to_remove = bddtrue;
   set<string>::const_iterator it = to_remove.begin();

   while(it != to_remove.end())
   {
     bdd_to_remove &= conaux::get_variable(*it);
     ++it;
   }

   constraint = bdd_exist(constraint, bdd_to_remove);
   
   set<string> new_cvars;
   set_intersection(cvars.begin(), cvars.end(), to_keep.begin(), to_keep.end(), inserter(new_cvars, new_cvars.end()));
   cvars = new_cvars;
}

std::set<std::string> con_type::get_cvars()
{
   return cvars;
}

ostream &operator<<(ostream &os, const con_type &con)
{
   os << "{ ";
   set<string>::iterator it = con.cvars.begin();
   while(it != con.cvars.end())
   {
      os << *it << "(" << bdd_var(conaux::get_variable(*it)) << ") ";
      ++it;
   }
   os << "}" << endl;
   os << bddtable << con.constraint << flush;
   return os;
}
