#ifndef LOG_H
#define LOG_H

#include <iostream>
#include <fstream>
#include <set>
#include <iterator>

/**
 * These are the log types possible to use.
 *
 * L_MODEL: log model
 *
 * L_REPORT: log reporting of answer
 *
 * L_STATUS: log main status messages
 *
 * L_RES: log resource usage messages
 *
 * L_DBG: log (unimportant) debugging messages
 *
 * L_ERR: log error messages
 *
 * L_EXT: log model extension messages
 */
enum log_type { L_MODEL = 1, L_REPORT, L_STATUS, L_RES, L_DBG, L_ERR, L_EXT };

/**
 * This class acts as a logger. It is assigned a set of log types and
 * an ostream to log to.
 */
class logger
{
public:

   /**
    * Write a text representation of this logger to an ostream.
    * \param os the ostream to write to.
    * \param l the logger to write.
    */
   friend std::ostream &operator<<(std::ostream &os, const logger &l);
     
   /**
    * Constructor taking the ostream to log to.
    * \param os the ostream.
    */
   logger(std::ostream &os);

   /**
    * Sets the log state of the logger (valid if the log type is
    * recognised, invalid otherwise) and returns the logger.  This
    * method should be called before every logging.
    * \param type the log type to use for the next logging.
    */
   logger &operator()(log_type type);

   /**
    * Log an element to the specified ostream (if the currently set
    * log type is valid) or discard the element (if the currently
    * set log type is invalid).
    * \param arg the element to log or discard.
    */
   template<class T>
   logger &operator<<(const T &arg);

   /**
    * Log an ostream manipulator to the specified ostream (if the
    * currently set log type is valid) or discard the element (if
    * the currently set log type is invalid).
    * \param pf the manipulator to log or discard.
    */
   logger &operator<<(std::ostream &(*pf)(std::ostream&));
      
   /**
    * Adds a log type to be recognised by this logger.
    * \param type the log type to be added.
    */
   void add_log_type(log_type type);

   /**
    * Log all members of the given container (if the currently set log
    * type is valid) or discard them (if the currently set log type
    * is invalid).
    * \param con the container whose members to log or discard.
    * \param delim the delimiter to log between each logged element.
    */
   template<class Con>
   void log_all(const Con &con, const std::string &delim);

   /**
    * Gets the log type associated with the given name.
    * \param name the log name.
    * \return The log type.
    */
   static log_type get_log_type(const std::string &name);

   /**
    * Gets the name associated with the given log type.
    * \param type the log type.
    * \return The log name.
    */
   static std::string get_log_name(const log_type &type);

private:

   /**
    * The ostream associated with this logger.
    */
   std::ostream &os;

   /**
    * The set of log types recognised by this logger.
    */
   std::set<log_type> log_types;

   /**
    * Is the logger valid? (Is a recognised log type set?)
    */
   bool logger_valid;
};

extern logger lout;
extern logger lerr;

/* These methods are defined in the headerfile because they are
   templated and cannot be in any file other than that in which they
   were declared. */
 
template<class T>
logger &logger::operator<<(const T &arg)
{
   if(logger_valid)
      os << arg;
   return *this;
}

template<class Con>
void logger::log_all(const Con &con, const std::string &delim)
{
   if(logger_valid)
      std::copy(con.begin(), con.end(), std::ostream_iterator<typename Con::value_type>(os, delim.c_str()));
}


#endif // LOG_H
