#include <iostream>
#include <iomanip>

#include "parser-classes.hpp"
#include "unify.hpp"
#include "constraint.hpp"
#include "compmod.hpp"
#include "log.hpp"
#include "resmes.hpp"

using namespace std;

extern int yyparse();
extern prolog_program *program;
extern int yynerrs;

long secdiff, usecdiff;      

void parse_step()
{
   lout(L_STATUS) << "Parsing... ";

   resources["parsing"].set_desc("Parsing");
   resources["parsing"].start();
   yyparse();
   resources["parsing"].stop();

   lout(L_STATUS) << "Parsed." << endl;

   if(yynerrs > 0) return;

   lout(L_DBG) << program << endl;
   
   n_base::delete_us(); // This is not necessary if parsing was successful
}

void live_step(const string &strats_gabc_and_local, string &random_seed)
{
   bool unfinished;
   c_rule *query = program->rules[program->rules.size() - 1];
   program->rules.erase(--program->rules.end(), program->rules.end());

   compmod::entry::init();

   compmod computation_model(query, random_seed);
   lout(L_STATUS) << "Random seed is " << random_seed << endl;

   istringstream iss(strats_gabc_and_local);
   size_t local;
   string strat_g, strat_a, strat_b, strat_c, strat_p;
   
   iss >> strat_g >> strat_a >> strat_b >> strat_c >> strat_p >> local;

   // Set the config values
   try { computation_model.set_pick_strategy(strat_g); }
   catch(const runtime_error &e)
   {
      lout(L_ERR) << e.what() << endl;
      return;
   }
   computation_model.set_ans_a_pick_strategy(compmod::get_strategy(strat_a));
   computation_model.set_ans_b_pick_strategy(compmod::get_strategy(strat_b));
   computation_model.set_calls_pick_strategy(compmod::get_strategy(strat_c));
   computation_model.set_possibles_strategy(compmod::get_strategy(strat_p));
   computation_model.set_local_ebc_percentage(local);

   lout(L_MODEL) << "\n\nStep 0\n----" << endl << computation_model << endl;

   try
   {
      int step = 0;
      while((unfinished = computation_model.extend_set()))
      {
	 lout(L_MODEL) << "\n\nStep " << (++step) << "\n----" << endl << computation_model << endl;
      }
   
      lout(L_STATUS) << "\n\n\nFixpoint calculation completed in " << step << " steps." << endl;

      computation_model.report_answer(query->head);
      lout(L_REPORT) << endl;
      
      /*
      set<compmod::entry>::const_iterator a, b, c, oa, ob, oc;
      a = computation_model.ans_a.begin();
      b = computation_model.ans_b.begin();
      c = computation_model.calls.begin();
      oa = computation_model.ob_ans_a.begin();
      ob = computation_model.ob_ans_b.begin();
      oc = computation_model.ob_calls.begin();
      cout << computation_model.ans_a.size() << " a" << endl;
      cout << computation_model.ans_b.size() << " b" << endl;
      cout << computation_model.calls.size() << " c" << endl;
      cout << computation_model.ob_ans_a.size() << " oa" << endl;
      cout << computation_model.ob_ans_b.size() << " ob" << endl;
      cout << computation_model.ob_calls.size() << " oc" << endl;

      cout << "ANS_A" << endl;
      while(a != computation_model.ans_a.end())
      {
         cout << a->rule << endl;
         ++a;
      }
      cout << "ANS_B" << endl;
      while(b != computation_model.ans_b.end())
      {
         cout << b->rule << endl;
         ++b;
      }
      cout << "CALLS" << endl;
      while(c != computation_model.calls.end())
      {
         cout << c->rule << endl;
         ++c;
      }
      cout << "OBANS_A" << endl;
      while(oa != computation_model.ob_ans_a.end())
      {
         cout << oa->rule << endl;
         ++oa;
      }
      cout << "OB_ANS_B" << endl;
      while(ob != computation_model.ob_ans_b.end())
      {
         cout << ob->rule << endl;
         ++ob;
      }
      cout << "OB_CALLS" << endl;
      while(oc != computation_model.ob_calls.end())
      {
         cout << oc->rule << endl;
         ++oc;
      }
      */
   }
   catch(compmod::compmod_error &e)
   {
      lerr(L_ERR) << "A fatal error occured in the computation model." << endl;
   }

   computation_model.finalise();

   query->deep_delete();
}

bool parse_config(istream &os, map<string, string> &cfg)
{
   // Default strategies
   cfg["GLOBAL_STRATEGY"] = "weighted";
   cfg["STRATEGY_ANS_A"] = cfg["STRATEGY_ANS_B"] = cfg["STRATEGY_CALLS"] 
      = cfg["STRATEGY_POSSIBLES"] = compmod::get_strategy("random");
   
   // Random seed
   cfg["RANDOM_SEED"] = "random";

   string line;
   while(getline(os, line))
   {
      istringstream iss(line);
      string word;
      
      if(iss >> word)
      {
	 // Skip comments
	 if(word[0] == '#')
	    continue;

	 if(word == "LOG")
	 {
	    while(iss >> word)
	    {
	       lout.add_log_type(logger::get_log_type(word));
	       lerr.add_log_type(logger::get_log_type(word));
	    }
	 }
	 else
	 {
	    static string config_tokens[] = { "PERCENT_LOCAL_EXTENSIONS", "GLOBAL_STRATEGY", "STRATEGY_ANS_A", 
					      "STRATEGY_ANS_B", "STRATEGY_CALLS", "STRATEGY_POSSIBLES",
					      "BUDDY_NODESIZE", "BUDDY_CACHESIZE", "CONSTRAINT_INIT_VARNUM",
                                              "RANDOM_SEED" };
	    
	    for(size_t i = 0; i < sizeof(config_tokens) / sizeof(string); i++)
	    {
	       if(word != config_tokens[i])
		  continue;

	       if(!(iss >> cfg[config_tokens[i]]))
		  return false;
	    }
	 }
      }
   }

   lout(L_STATUS) << "Configuration options are" << endl
		  << "-------------------------" << endl;
   map<string, string>::iterator it = cfg.begin();
   while(it != cfg.end())
   {
      lout(L_STATUS) << "   " << it->first << " = " << it->second << endl;
      ++it;
   }
   lout(L_STATUS) << "   " << lout << endl;
   lout(L_STATUS) << endl;

   // Config file parsing succeeded
   return true;
}

int main(int argc, char **argv)
{
   if(argc != 2)
   {
      cout << "Usage: " << argv[0] << " <config file>" << endl
	   << "(Program input is read from stdin.)" << endl;
      return 0;
   }

   // Parse the config file
   map<string, string> cfg;
   ifstream conffile(argv[1]);
   if(!conffile)
   {
      cout << "Config file could not be read." << endl;
      return 1;
   }
   else
   {
      if(!parse_config(conffile, cfg))
      {
	 cout << "Config file was not parsed successfully." << endl;
	 return 1;   
      }
   }
   
   // Save all pointers to c_base derived objects for deletion later.
   c_base::set_safety(true);

   // Initialise the constraint stuff
   conaux::init(cfg["BUDDY_NODESIZE"]+  " " + cfg["BUDDY_CACHESIZE"] + " " + cfg["CONSTRAINT_INIT_VARNUM"]);

   // Parse the program (including the query) from stdin
   parse_step();

   // We take care of our garbage ourselves from here on
   c_base::set_safety(false);

   if(yynerrs != 0)
   {
      lerr(L_ERR) << "An error occured during parsing." << endl;
      n_base::delete_us();
      c_base::delete_us();
   }
   else
   {
      resources["fixpoint"].set_desc("Fixpoint calculation");
      resources["fixpoint"].start();
      // Go through the fixpoint calculation
      live_step(cfg["GLOBAL_STRATEGY"] + " " + cfg["STRATEGY_ANS_A"] + " "
		+ cfg["STRATEGY_ANS_B"] + " " + cfg["STRATEGY_CALLS"] + " " 
		+ cfg["STRATEGY_POSSIBLES"] + " " + cfg["PERCENT_LOCAL_EXTENSIONS"],
                cfg["RANDOM_SEED"]);
      resources["fixpoint"].stop();
   }

   lout(L_RES) << endl
	       << "Resource statistics" << endl
	       << "-------------------" << endl;
   map<string, resmes>::iterator it = resources.begin();
   while(it != resources.end())
   {
      lout(L_RES) << "   " << it->second;
      ++it;
   }
   lout(L_RES) << endl;
	       
   lout(L_DBG) << endl
	       << "Memory debug" << endl
	       << "------------" << endl
	       << "n_* left: " << n_base::count << endl
	       << "c_terms left: " << c_term::count << endl
	       << "c_rules left: " << c_rule::count << endl
	       << "unifiers left: " << unifier::count << endl;

   lout(L_STATUS) << endl << "Random seed was " << cfg["RANDOM_SEED"] << endl;

   // Finalise the constraint stuff
   conaux::finalise();

   return 0;
}
