#include <set>
#include <iostream>
#include <sstream>
#include <algorithm>
#include <iterator>

#include "parser-classes.hpp"
#include "compmod.hpp"
#include "log.hpp"
#include "resmes.hpp"

using namespace std;

extern StringsHolder stringsholder;

int c_rule::global_variable_index = 0;
con_type n_body::constraint(true);

int c_term::count = 0;
int c_rule::count = 0;
int n_base::count = 0;

c_term::~c_term()
{
   count--;
   map<const c_term*, int>::const_iterator it = sorted_terms.begin();
   while(it != sorted_terms.end())
   {
      it->first->sorted_terms.erase(this);
      ++it;
   }
}

c_rule::~c_rule()
{
   count--;
   map<const c_rule*, int>::const_iterator it = sorted_rules.begin();
   while(it != sorted_rules.end())
   {
      it->first->sorted_rules.erase(this);
      ++it;
   }
}

void c_rule::sort_before(const c_rule *r) const
{
   sorted_rules[r] = -1;
   r->sorted_rules[this] = 1;
}

void c_rule::sort_same(const c_rule *r) const
{
   sorted_rules[r] = r->sorted_rules[this] = 0;
}

n_base::~n_base()
{
   count--;
}

ostream &operator<<(ostream &os, const c_term *t)
{
   switch(t->type)
   {
      case T_VARIABLE:
	 os << t->name;
	 break;
      case T_FUNCTOR:
	 if(t->negated)
	    os << "\\+";
	 os << t->name;
	 if(t->arglist.size() == 0)
	    return os;
	 os << "(";
	 for(size_t i = 0; i < t->arglist.size(); i++)
	 {
	    if(i != 0) os << ", ";
	    os << t->arglist[i];
	 }
	 os << ")";
	 break;
      case T_LIST:
	 os << "[";
	 while(!t->list_empty)
	 {
	    os << t->list_head;	    
	    if(!t->list_tail->list_empty)
	       os << ",";
	    else if(t->list_tail->match_rest)
	    {
	       os << "|";
	       os << t->list_tail;
	    }
	    t = t->list_tail;
	 }
	 os << "]";
	 break;
      default:
	 os << "?" << t->type;
   }

   return os;
}

ostream &operator<<(ostream &os, const c_rule *r)
{
   os << r->head;

   if(r->body.size() != 0)
   {
      os << " <- ";
      os << r->body[0];
   }

   for(size_t i = 1; i < r->body.size(); i++)
   {
      os << ", ";
      os << r->body[i];
   }

   os << "." << flush;

   return os;
}

ostream &operator<<(ostream &os, const prolog_program *p)
{
   for(size_t i = 0; i < p->rules.size(); i++)
   {
      os << p->rules[i];
      os << endl;
   }

   return os;
}

vector<n_base*> n_base::nodes;

void n_base::delete_us()
{
   for(size_t i = 0; i < nodes.size(); i++)
      delete nodes[i];

   nodes.clear();
}

n_base::n_base()
{
   count++;
   nodes.push_back(this);
}

bool c_base::safety;
vector<c_base*> c_base::c_elements;

c_base::~c_base() {}

void c_base::set_safety(const bool safe)
{
   safety = safe;
}

void c_base::delete_us()
{
   for(size_t i = 0; i < c_elements.size(); i++)
      delete c_elements[i];

   c_elements.clear();
}

c_base::c_base()
{
   if(safety)
      c_elements.push_back(this);
}

c_rule::c_rule(): head(new c_term)
{
   count++;
}

// static
bool c_sorter::c_rule_lt(const c_rule *c1, const c_rule *c2)
{
   if(c1 == c2) return false;
   
   if((c1->body.size() == 0 && c2->body.size() != 0)
      || (c1->body.size() != 0 && c2->body.size() == 0))
   {
      lerr(L_ERR) << "Calls and answers must not be mixed in the same set!" << endl;
      throw compmod::compmod_error();
   }

   return c_rule_order(c1, c2) == -1;
}

// static
int c_sorter::c_rule_order(const c_rule *c1, const c_rule *c2)
{
   map<const c_rule*, int>::const_iterator it = c1->sorted_rules.find(c2);
   if(it != c1->sorted_rules.end())
      return it->second;

   switch(c1->body.size())
   {
      case 0: // This is a set of answers
      {
         // Sort on the head
         int order = c_term_order(c1->head, c2->head);
         switch(order)
         {
            case -1:
               c1->sort_before(c2);
               return -1;
               break;
            case 1:
               c2->sort_before(c1);
               return 1;
               break;
            default:
               c1->sort_same(c2);
               return 0;
               break;
         }
      }
      default: // This is a set of calls
      {
         // Try differentiating by body literals
         for(size_t i = 0; i < c1->body.size() && i < c2->body.size(); i++)
         {
            int order2 = c_term_order(c1->body[i], c2->body[i]);
            switch(order2)
            {
               case -1:
                  c1->sort_before(c2);
                  return -1;
                  break;
               case 1:
                  c2->sort_before(c1);
                  return 1;
                  break;
               default:
                  break;
            }
         }
         
         // Try differentiating by body length
         if(c1->body.size() < c2->body.size())
         {
            c1->sort_before(c2);
            return -1;
         }
         else if(c2->body.size() < c1->body.size())
         {
            c2->sort_before(c1);
            return 1;
         }
         
         // Try differentiating by head
         int order2 = c_term_order(c1->head, c2->head);
         switch(order2)
         {
            case -1:
               c1->sort_before(c2);
               return -1;
               break;
            case 1:
               c2->sort_before(c1);
               return 1;
               break;
            default:
               // They are the same
               c1->sort_same(c2);
               return 0;
         }
      }   
   }
}

void c_rule::deep_delete()
{
   head->deep_delete();
   
   for(size_t i = 0; i < body.size(); i++)
      body[i]->deep_delete();

   delete this;
}

void c_rule::deep_copy(c_rule *copy, const bool reindex) const
{
   // Are we instantiating? If so, increase the index
   if(reindex) global_variable_index++;

   // Copy the constraint and constraint variables
   copy->constraint = constraint;

   // If we are reindexing the variables, do so for the constraint variables
   if(reindex)
      conaux::reindex(copy->constraint);

   // Copy the head
   head->deep_copy(copy->head, reindex);

   // Copy the body
   c_term *temp;
   for(size_t i = 0; i < body.size(); i++)
   {
      temp = new c_term;
      body[i]->deep_copy(temp, reindex);
      copy->body.push_back(temp);
   }
}

void c_term::sort_before(const c_term *t) const
{
   sorted_terms[t] = -1;
   t->sorted_terms[this] = 1;
}

void c_term::sort_same(const c_term *r) const
{
   sorted_terms[r] = r->sorted_terms[this] = 0;
}

void c_term::get_variables(set<string> &vars) const
{
   switch(type)
   {
      case T_VARIABLE:
	 vars.insert(name);
	 break;
	 
      case T_FUNCTOR:
	 for(size_t i = 0; i < arglist.size(); i++)
	    arglist[i]->get_variables(vars);
	 break;

      case T_LIST:
	 if(list_head != 0) list_head->get_variables(vars);
	 if(list_tail != 0) list_tail->get_variables(vars);
	 break;

      default:
	 break;
   }
}

void c_rule::get_variables(set<string> &vars) const
{
   head->get_variables(vars);
   for(size_t i = 0; i < body.size(); i++)
      body[i]->get_variables(vars);
}

void c_rule::project_constraint() const
{
   lout(L_DBG) << "Projecting constraint in: " << compmod::entry(this) << endl;

   resources["conproj"].start();

   set<string> existing_vars;
   get_variables(existing_vars);

   constraint.project(existing_vars);

   resources["conproj"].stop();

   lout(L_DBG) << "Result of projection: " << compmod::entry(this) << endl;
}

prolog_program::prolog_program(const vector<c_rule*> &rs)
{
   c_base::set_safety(false);

   // Copy the rules to our program, create new pointers
   c_rule *temp;
   for(size_t i = 0; i < rs.size(); i++)
   {
      temp = new c_rule;
      rs[i]->deep_copy(temp);
      temp->project_constraint();
      rules.push_back(temp);
   }

   // Delete the nodes and the strings that have been created during parsing
   n_base::delete_us();
   c_base::delete_us();
   stringsholder.delete_strings();
}

c_term::c_term(): list_head(0), list_tail(0),
		  list_empty(false), match_rest(false),
		  type(0)
{
   count++;
}

// static
bool c_sorter::c_term_lt(const c_term *c1, const c_term *c2)
{
   if(c1 == c2) return false;

   return c_term_order(c1, c2) == -1;
}

// static
int c_sorter::c_term_order(const c_term *c1, const c_term *c2)
{
   map<const c_term*, int>::const_iterator it = c1->sorted_terms.find(c2);
   if(it != c1->sorted_terms.end())
      return it->second;

   // Different types?
   if(c1->type < c2->type)
   {
      c1->sort_before(c2);
      return -1;
   }
   else if(c2->type < c1->type)
   {
      c2->sort_before(c1);
      return 1;
   }

   switch(c1->type) // Same as c2->type
   {
      case T_FUNCTOR:
      {
	 if(c1->name < c2->name)
         {
            c1->sort_before(c2);
            return -1;
         }
	 else if(c2->name < c1->name)
         {
            c2->sort_before(c1);
	    return 1;
         }

	 // Ok, same name - check arguments (same number of arguments)
	 
	 for(size_t i = 0; i < c1->arglist.size(); i++)
	 {
            int order = c_term_order(c1->arglist[i], c2->arglist[i]);
            switch(order)
            {
               case -1:
                  c1->sort_before(c2);
                  return -1;
                  break;
               case 1:
                  c2->sort_before(c1);
                  return 1;
                  break;
               default:
                  break;
            }
	 }
	 
	 // Separate negated and non-negated
	 
	 if(c1->negated && !c2->negated)
         {
            c1->sort_before(c2);
	    return -1;
         }
	 else if(!c1->negated && c2->negated)
         {
            c2->sort_before(c1);
	    return 1;
         }

         // They are the same
         c1->sort_same(c2);
	 return 0;
      }
      break;
      
      case T_LIST:
      {
	 if(c1->list_empty && !c2->list_empty)
         {
            c1->sort_before(c2);
	    return -1;
         }
	 else if(!c1->list_empty && c2->list_empty)
         {
            c2->sort_before(c1);
	    return 1;
	 }

	 // Make sure that we don't use 0:ed pointers
	 if(c1->list_head == 0 && c2->list_head == 0)
         {
            // They are the same
            c1->sort_same(c2);
	    return 0;
         }
	 else if(c1->list_head == 0)
         {
            c2->sort_before(c1);
	    return 1;
         }
	 else if(c2->list_head == 0)
         {
            c1->sort_before(c2);
	    return -1;
         }
	 
         int order = c_term_order(c1->list_head, c2->list_head);
         switch(order)
         {
            case -1:
               c1->sort_before(c2);
               return -1;
               break;
            case 1:
               c2->sort_before(c1);
               return 1;
               break;
            default:
               // Make sure that we don't use 0:ed pointers
               if(c1->list_tail == 0 && c2->list_tail == 0)
               {
                  // They are the same
                  c1->sort_same(c2);
                  return 0;
               }
               else if(c1->list_tail == 0)
               {
                  c2->sort_before(c1);
                  return 1;
               }
               else if(c2->list_tail == 0)
               {
                  c1->sort_before(c2);
                  return -1;
               }

               int order2 = c_term_order(c1->list_tail, c2->list_tail);
               switch(order2)
               {
                  case -1:
                     c1->sort_before(c2);
                     return -1;
                     break;
                  case 1:
                     c2->sort_before(c1);
                     return 1;
                     break;
                  default:
                     // They are the same
                     c1->sort_same(c2);
                     return 0;
                     break;
               }
	 }
      }
      break;
            
      case T_VARIABLE:
      default:
      {
         c1->sort_same(c2);
         return 0; // Variables are always equal
      }
      break;
   }
}

void n_functor::negate()
{
   if(term->negated)
      term->negated = false;
   else
      term->negated = true;
}

void c_term::deep_delete()
{
   for(size_t i = 0; i < arglist.size(); i++)
      arglist[i]->deep_delete();

   if(list_head != 0)
      list_head->deep_delete();
   if(list_tail != 0)
      list_tail->deep_delete();

   delete this;
}

void to_string(const int &val, ostringstream &oss)
{
   oss << val;
}

void c_term::deep_copy(c_term *copy, const bool reindex) const
{
   // Copy the type
   copy->type = type;

   // Copy the name
   copy->name = name;
   
   if(reindex && type == T_VARIABLE) // We are instancing a rule, use unique variable names
   {
      ostringstream oss;
      oss << "#" << c_rule::global_variable_index;
      copy->name += oss.str();
   }

   // Copy the arglist
   c_term *temp;
   for(size_t i = 0; i < arglist.size(); i++)
   {
      temp = new c_term;
      arglist[i]->deep_copy(temp, reindex);
      copy->arglist.push_back(temp);
   }

   // Copy the negation value
   copy->negated = negated;

   // Copy the list
   if(list_head != 0)
   {
      copy->list_head = new c_term;
      list_head->deep_copy(copy->list_head, reindex);
   }
   if(list_tail != 0)
   {
      copy->list_tail = new c_term;
      list_tail->deep_copy(copy->list_tail, reindex);      
   }
   copy->list_empty = list_empty;
   copy->match_rest = match_rest;
}

void c_term::insert_list(const vector<c_term*> &list,
			 const c_term *rest, 
			 const size_t start_index)
{
   // Every element in the sequence is a linked list
   type = T_LIST;

   // The list is empty
   if(start_index >= list.size())
   {
      if(rest == 0)
      {
	 // Show that this is the empty list
	 list_empty = true;
      }
      else if(rest->type == T_LIST)
      {
	 // To put a list after a pipe is the same as continuing the list
	 rest->deep_copy(this);
      }
      else
      {
	 // Copy the element representing the rest of the list here
	 rest->deep_copy(this);
	 
         // Show that this matches the rest of the list
	 match_rest = true;
	 list_empty = true;

	 // Show that there are no more elements
	 list_tail = new c_term;
	 list_tail->list_empty = true;
      }
      return;
   }
   else
   {
      // Set the head element
      list_head = list[start_index];
      // Create the tailing list
      list_tail = new c_term;
      // Insert the rest of the list into the tail
      list_tail->insert_list(list, rest, start_index + 1);
   }  
}

n_rules::n_rules(const n_rule *r, const n_rules *rs)
{
   rules.push_back(r->rule);

   if(rs == 0) return;

   rules.insert(rules.end(), rs->rules.begin(), rs->rules.end());
}

n_rule::n_rule(const n_functor *f, const n_body *b):
   rule(new c_rule)
{
   rule->head = f->term;

   // Copy the values aquired during parsing of the body
   rule->constraint = n_body::constraint;
   n_body::constraint = true;

   if(b == 0) return;

   rule->body = b->functors;
}

n_rule::n_rule(const n_functor *f): rule(new c_rule)
{
   rule->head = f->term;
   rule->constraint = true;
}

n_body::n_body(const n_functor *f, const n_body *b)
{
   if(f->term->name == "sat")
      constraint = conaux::resolve_sat(f->term);
   else
      functors.push_back(f->term);
   
   if(b == 0) return;

   functors.insert(functors.end(), b->functors.begin(), b->functors.end());
}

n_arg::n_arg(): term(new c_term) {}

n_arglist::n_arglist(const n_arg *a,
		     const n_arglist *al)
{
   terms.push_back(a->term);

   if(al == 0) return;

   terms.insert(terms.end(), al->terms.begin(), al->terms.end());
}

n_functor::n_functor(const string n, const n_arglist *al)
{
   term->name = n;
   term->negated = false;
   term->type = T_FUNCTOR;

   if(al == 0) return;

   term->arglist = al->terms;
}

n_variable::n_variable(string n)
{
   term->name = n;
   term->type = T_VARIABLE;
}

n_list::n_list(const n_arglist *al)
{
   term->insert_list(al->terms);
}

n_list::n_list(const n_arglist *left_al, 
	       const n_arg *right_a)
{
   term->insert_list(left_al->terms, right_a->term);
}

n_list::n_list()
{
   vector<c_term*> vec; // empty list
   term->insert_list(vec);
}
